@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
+# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -13,7 +13,7 @@
 ##############################################################################
 """Database connection support
 
-$Id: Connection.py 106502 2009-12-14 19:27:08Z jim $"""
+$Id: Connection.py 117033 2010-09-28 20:27:49Z jim $"""
 
 import logging
 import sys
@@ -262,13 +262,17 @@ class Connection(ExportImport, object):
         return obj
 
     def cacheMinimize(self):
-        """Deactivate all unmodified objects in the cache."""
-        self._cache.minimize()
+        """Deactivate all unmodified objects in the cache.
+        """
+        for connection in self.connections.itervalues():
+            connection._cache.minimize()
 
     # TODO: we should test what happens when cacheGC is called mid-transaction.
     def cacheGC(self):
-        """Reduce cache size to target size."""
-        self._cache.incrgc()
+        """Reduce cache size to target size.
+        """
+        for connection in self.connections.itervalues():
+            connection._cache.incrgc()
 
     __onCloseCallbacks = None
     def onCloseCallback(self, f):
@@ -335,13 +339,13 @@ class Connection(ExportImport, object):
     def invalidate(self, tid, oids):
         """Notify the Connection that transaction 'tid' invalidated oids."""
         if self.before is not None:
-            # this is an historical connection.  Invalidations are irrelevant.
+            # This is a historical connection.  Invalidations are irrelevant.
             return
         self._inv_lock.acquire()
         try:
             if self._txn_time is None:
                 self._txn_time = tid
-            elif tid < self._txn_time:
+            elif (tid < self._txn_time) and (tid is not None):
                 raise AssertionError("invalidations out of order, %r < %r"
                                      % (tid, self._txn_time))
 
@@ -444,8 +448,12 @@ class Connection(ExportImport, object):
             assert oid is not None
             if oid in self._added:
                 del self._added[oid]
+                if self._cache.get(oid) is not None:
+                    del self._cache[oid]
                 del obj._p_jar
                 del obj._p_oid
+                if obj._p_changed:
+                    obj._p_changed = False
             else:
 
                 # Note: If we invalidate a non-ghostifiable object
@@ -741,6 +749,8 @@ class Connection(ExportImport, object):
         self._invalidate_creating()
         while self._added:
             oid, obj = self._added.popitem()
+            if obj._p_changed:
+                obj._p_changed = False
             del obj._p_oid
             del obj._p_jar
         self._tpc_cleanup()
@@ -749,15 +759,18 @@ class Connection(ExportImport, object):
         """Disown any objects newly saved in an uncommitted transaction."""
         if creating is None:
             creating = self._creating
-            self._creating = {}
 
         for oid in creating:
             o = self._cache.get(oid)
             if o is not None:
                 del self._cache[oid]
+                if o._p_changed:
+                    o._p_changed = False
                 del o._p_jar
                 del o._p_oid
 
+        creating.clear()
+
     def tpc_vote(self, transaction):
         """Verify that a data manager can commit the transaction."""
         try:
@@ -1130,7 +1143,10 @@ class Connection(ExportImport, object):
         self._creating.clear()
         self._registered_objects = []
 
-        state = self._storage.position, self._storage.index.copy()
+        state = (self._storage.position,
+                 self._storage.index.copy(),
+                 self._storage.creating.copy(),
+                 )
         result = Savepoint(self, state)
         # While the interface doesn't guarantee this, savepoints are
         # sometimes used just to "break up" very long transactions, and as
@@ -1143,8 +1159,10 @@ class Connection(ExportImport, object):
         self._abort()
         self._registered_objects = []
         src = self._storage
-        self._cache.invalidate(src.index)
+        self._invalidate_creating(src.creating)
+        index = src.index
         src.reset(*state)
+        self._cache.invalidate(index)
 
     def _commit_savepoint(self, transaction):
         """Commit all changes made in savepoints and begin 2-phase commit
@@ -1153,38 +1171,43 @@ class Connection(ExportImport, object):
         self._storage = self._normal_storage
         self._savepoint_storage = None
 
-        self._log.debug("Committing savepoints of size %s", src.getSize())
-        oids = src.index.keys()
+        try:
+            self._log.debug("Committing savepoints of size %s", src.getSize())
+            oids = src.index.keys()
 
-        # Copy invalidating and creating info from temporary storage:
-        self._modified.extend(oids)
-        self._creating.update(src.creating)
+            # Copy invalidating and creating info from temporary storage:
+            self._modified.extend(oids)
+            self._creating.update(src.creating)
 
-        for oid in oids:
-            data, serial = src.load(oid, src)
-            obj = self._cache.get(oid, None)
-            if obj is not None:
-                self._cache.update_object_size_estimation(obj._p_oid, len(data))
-                obj._p_estimated_size = len(data)
-            if isinstance(self._reader.getGhost(data), Blob):
-                blobfilename = src.loadBlob(oid, serial)
-                s = self._storage.storeBlob(oid, serial, data, blobfilename,
+            for oid in oids:
+                data, serial = src.load(oid, src)
+                obj = self._cache.get(oid, None)
+                if obj is not None:
+                    self._cache.update_object_size_estimation(
+                        obj._p_oid, len(data))
+                    obj._p_estimated_size = len(data)
+                if isinstance(self._reader.getGhost(data), Blob):
+                    blobfilename = src.loadBlob(oid, serial)
+                    s = self._storage.storeBlob(
+                        oid, serial, data, blobfilename,
+                        '', transaction)
+                    # we invalidate the object here in order to ensure
+                    # that that the next attribute access of its name
+                    # unghostify it, which will cause its blob data
+                    # to be reattached "cleanly"
+                    self.invalidate(None, (oid, ))
+                else:
+                    s = self._storage.store(oid, serial, data,
                                             '', transaction)
-                # we invalidate the object here in order to ensure
-                # that that the next attribute access of its name
-                # unghostify it, which will cause its blob data
-                # to be reattached "cleanly"
-                self.invalidate(s, {oid:True})
-            else:
-                s = self._storage.store(oid, serial, data,
-                                        '', transaction)
 
-            self._handle_serial(s, oid, change=False)
-        src.close()
+                self._handle_serial(s, oid, change=False)
+        finally:
+            src.close()
 
     def _abort_savepoint(self):
         """Discard all savepoint data."""
         src = self._savepoint_storage
+        self._invalidate_creating(src.creating)
         self._storage = self._normal_storage
         self._savepoint_storage = None
 
@@ -1204,8 +1227,11 @@ class Connection(ExportImport, object):
         # by another thread, so the risk of a reread is pretty low.
         # It's really not worth the effort to pursue this.
 
+        # Note that we do this *after* reseting the storage so that, if
+        # data are read, we read it from the reset storage!
+
         self._cache.invalidate(src.index)
-        self._invalidate_creating(src.creating)
+
         src.close()
 
     # Savepoint support
@@ -1333,7 +1359,7 @@ class TmpStore:
     def temporaryDirectory(self):
         return self._storage.temporaryDirectory()
 
-    def reset(self, position, index):
+    def reset(self, position, index, creating):
         self._file.truncate(position)
         self.position = position
         # Caution:  We're typically called as part of a savepoint rollback.
@@ -1346,6 +1372,7 @@ class TmpStore:
         # a copy of the index here.  An alternative would be to ensure that
         # all callers pass copies.  As is, our callers do not make copies.
         self.index = index.copy()
+        self.creating = creating
 
 class RootConvenience(object):
 
